module Algebra (Vector(..), IPVector(..), NormVector(..), Algebra(..), AlgebraU(..)) where

infixl 7 @*, .*, ./
infixl 6 @+, @-

class Vector v where
	zvec :: v
	(.*) :: v -> Double -> v
	(./) :: v -> Double -> v
	(@+) :: v -> v -> v
	(@-) :: v -> v -> v
	a @- b = a @+ b .* (-1)
	a ./ b = a .* (1 / b)

class (NormVector v) => IPVector v where
	(@.) :: v -> v -> Double
	-- norm should be the one derived from the inner product (@.)

class (Vector v) => NormVector v where
	norm :: v -> Double
	norm = sqrt . normSqr
	normSqr :: v -> Double
	normSqr = (^ 2) . norm
	normalize :: v -> v
	normalize x = x ./ norm x

class (Vector v) => Algebra v where
	(@*) :: v -> v -> v

class (Algebra v) => AlgebraU v where
	idvec :: v -- identity
	vprod :: [v] -> v
	vprod = foldl (@*) idvec

instance Vector Double where
	zvec = 0
	(.*) = (*)
	(@+) = (+)

instance Vector () where
	zvec = ()
	_ .* _ = ()
	_ @+ _ = ()

instance (Vector v0, Vector v1) => Vector (v0, v1) where
	zvec = (zvec, zvec)
	(x0, x1) .* a = (x0 .* a, x1 .* a)
	(x0, x1) @+ (y0, y1) = (x0 @+ y0, x1 @+ y1)

instance (Vector v0, Vector v1, Vector v2) => Vector (v0, v1, v2) where
	zvec = (zvec, zvec, zvec)
	(x0, x1, x2) .* a = (x0 .* a, x1 .* a, x2 .* a)
	(x0, x1, x2) @+ (y0, y1, y2) = (x0 @+ y0, x1 @+ y1, x2 @+ y2)

instance (Vector v) => Vector (a -> v) where
	zvec = const zvec
	(u @+ v) x = u x @+ v x
	(u .* c) x = u x .* c

instance IPVector Double where
	(@.) = (*)

instance IPVector () where
	_ @. _ = 0

instance (IPVector v0, IPVector v1) => IPVector (v0, v1) where
	(x0, x1) @. (y0, y1) = x0 @. y0 + x1 @. y1

instance (IPVector v0, IPVector v1, IPVector v2) => IPVector (v0, v1, v2) where
	(x0, x1, x2) @. (y0, y1, y2) = x0 @. y0 + x1 @. y1 + x2 @. y2

instance NormVector Double where
	norm = Prelude.id

instance NormVector () where
	normSqr = const 0

instance (IPVector v0, IPVector v1) => NormVector (v0, v1) where
	normSqr (v0, v1) = normSqr v0 + normSqr v1

instance (IPVector v0, IPVector v1, IPVector v2) => NormVector (v0, v1, v2) where
	normSqr (v0, v1, v2) = normSqr v0 + normSqr v1 + normSqr v2

instance Algebra Double where
	(@*) = (*)

instance (Algebra v) => Algebra (a -> v) where
	(u @* v) x = u x @* v x

instance AlgebraU Double where
	idvec = 1

instance (AlgebraU v) => AlgebraU (a -> v) where
	idvec = const idvec
