{-# OPTIONS -funbox-strict-fields #-}

module Color (Color, RGB(..), LongRGB(..), white, rgbToLongRGB) where

import Algebra

import Data.Monoid

import Foreign.Ptr
import Foreign.Storable

import Geometry
import Types

type Color = RGB
data RGB = RGB !Double !Double !Double deriving Show
data LongRGB = LongRGB !LongAccum !LongAccum !LongAccum

instance Storable RGB where
	sizeOf _ = 3 * sizeOf (undefined :: Double)
	alignment _ = alignment (undefined :: Double)
	peekElemOff address index =
		do	r <- p (3 * index)
			g <- p (3 * index + 1)
			b <- p (3 * index + 2)
			return (RGB r g b)
		where
			p = peekElemOff (castPtr address)
	pokeElemOff address index (RGB r g b) =
		do	p (index * 3) r
			p (index * 3 + 1) g
			p (index * 3 + 2) b
		where
			p = pokeElemOff (castPtr address)
	peek = flip peekElemOff 0
	poke = flip pokeElemOff 0

instance Storable LongRGB where
	sizeOf _ = 3 * sizeOf (undefined :: LongAccum)
	alignment _ = alignment (undefined :: LongAccum)
	peekElemOff address index =
		do	r <- p (3 * index)
			g <- p (3 * index + 1)
			b <- p (3 * index + 2)
			return (LongRGB r g b)
		where
			p = peekElemOff (castPtr address)
	pokeElemOff address index (LongRGB r g b) =
		do	p (index * 3) r
			p (index * 3 + 1) g
			p (index * 3 + 2) b
		where
			p = pokeElemOff (castPtr address)
	peek = flip peekElemOff 0
	poke = flip pokeElemOff 0

instance Monoid LongRGB where
    -- component-by-component addition
    mempty = LongRGB mempty mempty mempty
    mappend (LongRGB r0 g0 b0) (LongRGB r1 g1 b1) = LongRGB (mappend r0 r1) (mappend g0 g1) (mappend b0 b1)

rgbToLongRGB :: RGB -> LongRGB
rgbToLongRGB (RGB r g b) = LongRGB (doubleToLA r) (doubleToLA g) (doubleToLA b)

white :: Color
white = RGB 1 1 1

instance Vector RGB where
	zvec = RGB 0 0 0
	(RGB r0 g0 b0) @+ (RGB r1 g1 b1) = RGB (r0 + r1) (g0 + g1) (b0 + b1)
	(RGB r g b) .* c = RGB (r * c) (g * c) (b * c)

instance NormVector RGB where
	norm (RGB r g b) = r + g + b

instance Algebra RGB where
	(RGB r0 g0 b0) @* (RGB r1 g1 b1) = RGB (r0 * r1) (g0 * g1) (b0 * b1)

instance AlgebraU RGB where
	idvec = RGB 1 1 1
