module Display (Display.init, display, run_with_glut) where

import Algebra
import Control.Concurrent
import Data.IORef
import Geometry
import GlobalInfo

import Foreign.C.Types
import Foreign.Marshal.Array
import Foreign.Ptr

import Graphics.Rendering.OpenGL.GL.BasicTypes
import Graphics.Rendering.OpenGL.GL.BeginEnd
import Graphics.Rendering.OpenGL.GL.CoordTrans
import qualified Graphics.Rendering.OpenGL.GL.PixelRectangles.Rasterization as Rasterization
import Graphics.Rendering.OpenGL.GL.StateVar
import Graphics.Rendering.OpenGL.GL.VertexArrays
import Graphics.Rendering.OpenGL.GL.VertexSpec
import Graphics.Rendering.OpenGL.GLU.Matrix

import Graphics.UI.GLUT.Initialization
import Graphics.UI.GLUT.Callbacks.Global
import Graphics.UI.GLUT.Callbacks.Window
import Graphics.UI.GLUT.Window

import IO

import System
import System.Exit

import TheScene

-- TODO: try making Vec a newtype of Vector3 or Vertex3 for speed
vecToVec3 :: Vec -> Vector3 Double
vecToVec3 (V x y z) = Vector3 x y z

vecToVer3 :: Vec -> Vertex3 Double
vecToVer3 (V x y z) = Vertex3 x y z

init :: GlobalInfo -> IO ()
init global =
	do	getArgs >>= initialize "MLT Raytracer"
		initialDisplayMode $= [RGBMode, DoubleBuffered]
		initialWindowSize $= (Size 128 128)
		createWindow "dummy window"
		displayCallback $= return ()
		idleCallback $= Just (readIORef (exit global) >>= \e -> if e then exitWith ExitSuccess else yield)

display :: Int -> Int -> Ptr CUChar -> IORef [Point] -> IO ()
display width height img_data path_ref =
	do	Rasterization.drawPixels size (Rasterization.PixelData Rasterization.RGB UnsignedByte img_data)
		path <- readIORef path_ref
		matrixMode $= Projection
		loadIdentity
		frustum (-norm right / 10) (norm right / 10) (-norm up / 10) (norm up / 10) (norm gaze / 10) (norm gaze * 100)
		lookAt (vecToVer3 eye) (vecToVer3 (eye @+ gaze)) (vecToVec3 up)
		color (Color3 0 1 1 :: Color3 Double)
		renderPrimitive LineStrip (sequence_ [vertex (Vertex3 x y z) | V x y z <- path])
		swapBuffers
	where
		size = Size (fromIntegral width) (fromIntegral height)

run_with_glut :: GlobalInfo -> IO () -> IO ()
run_with_glut = const Prelude.id
