module Image (nplot, image_norm, image_bytes, write_PPM, longImageToImage) where

import Algebra
import Color

import Data.Array.IArray
import Data.Array.MArray
import Data.Monoid

import Foreign.C.Types
import Foreign.Marshal.Array

import System.IO

import Types

nplot :: (MArray a c m, Monoid c) => a (Int, Int) c -> c -> Double -> Double -> m ()
nplot image c x y =
	do	((x0, y0), (x1, y1)) <- getBounds image
		let	x' = floor (fromIntegral x0 + fromIntegral (x1 - x0) * (x + 1) / 2)
			y' = floor (fromIntegral y0 + fromIntegral (y1 - y0) * (y + 1) / 2) in
			do	v <- readArray image (x', y')
				writeArray image (x', y') $! (mappend v c)

image_norm :: (IArray a c, NormVector c) => a (Int, Int) c -> Double
image_norm image = sqrt (sum (map normSqr (elems image)) / fromIntegral (width * height))
	where
		((x0, y0), (x1, y1)) = bounds image
		width = x1 - x0 + 1
		height = y1 - y0 + 1

image_bytes :: (IArray a RGB) => Bool -> a (Int, Int) RGB -> [CUChar]
image_bytes rev image = concat [rgbToBytes (image ! (x, y)) | y <- rows, x <- [x0..x1]]
	where
		rgbToBytes (RGB r g b) = map f [r, g, b]
			where
				f x = floor (min 255 (x * 256 / image_norm image))
		rows = if rev then [y1,pred y1..y0] else [y0..y1]
		((x0, y0), (x1, y1)) = bounds image

write_PPM :: (IArray a RGB) => Handle -> a (Int, Int) RGB -> IO ()
write_PPM file image =
	do	hSeek file AbsoluteSeek 0
		hPutStrLn file "P6"
		hPutStrLn file (show width ++ ' ' : show height ++ ' ' : "255")
		withArray (image_bytes True image) (flip (hPutBuf file) img_len)
	where
		((x0, y0), (x1, y1)) = bounds image
		width = x1 - x0 + 1
		height = y1 - y0 + 1
		img_len = 3 * width * height

longImageToImage :: (IArray a RGB, IArray a LongRGB) => a (Int, Int) LongRGB -> a (Int, Int) RGB
longImageToImage a = listArray bnds (map onePixel (elems a))
    where bnds = bounds a
          onePixel (LongRGB r g b) = RGB (oneNum r) (oneNum g) (oneNum b)
          oneNum x = mul * laToDouble x
          mul = 1 / maxVal
          maxVal = maximum [(laToDouble r) `max` (laToDouble g) `max` (laToDouble b) | LongRGB r g b <- elems a]
