module JobDescription
( JobDescription(..)
, readDesc
) where

import Color
import qualified Data.Map as Map
import Geometry
import qualified LightProp as LP
import Part
import Rand
import Metropolis
import Monad
import qualified Text.SExpression as SE
import Text.SExpression.Format
import TheScenes

data JobDescription =
    JobDescription
    { renderVersion :: String
    , scene :: forall g. RandomGenD g => Scene g
    , sampleMethod :: String
    , nReject :: Integer
    , imageWidth, imageHeight :: Int
    }

instance SExprFormat JobDescription where
    sExprFormat =
      ( 0
      , [ ( "render_version"
          , Field {fOptional = False, fGet = \jd -> Just [SE.Atom (renderVersion jd)]}
          )
        , ( "n_samples_reject"
          , Field {fOptional = False, fGet = \jd -> Just [SE.Atom (show (nReject jd))]}
          )
        , ( "sample_method"
          , Field {fOptional = False, fGet = \jd -> Just [SE.Atom (show (sampleMethod jd))]}
          )
        , ( "image_width"
          , Field {fOptional = False, fGet = \jd -> Just [SE.Atom (show (imageWidth jd))]}
          )
        , ( "image_height"
          , Field {fOptional = False, fGet = \jd -> Just [SE.Atom (show (imageHeight jd))]}
          )
        ]
      )
    getInitialFields = const []
    consFromFormat [] attribs =
        do rv <- getAtom "render_version" attribs
           sampm <- getAtom "sample_method" attribs
           nsr <- getAtom "n_samples_reject" attribs >>= readM
           iw <- getAtom "image_width" attribs >>= readM
           ih <- getAtom "image_height" attribs >>= readM
           return $
               JobDescription
               { renderVersion = rv
               , scene = the_scenes "box"
               , sampleMethod = sampm
               , nReject = nsr
               , imageWidth = iw
               , imageHeight = ih
               }

getAtom :: Monad m => String -> Map.Map String [SE.Value] -> m String
getAtom k m =
    case Map.lookup k m of
        Just [SE.Atom a] -> return a
        _ -> fail "JobDescription.getAtom: not a single atom"

readM :: (Monad m, Read a) => String -> m a
readM s = case reads s of [(x, [])] -> return x
                          _ -> fail "readM: failure"

readDesc :: Monad m => String -> m JobDescription
readDesc s =
    do r <- SE.readSExpr s
       case r of
           (SE.List attribs, []) -> readFormatted attribs
           _ -> fail "readDesc: bad s-expression"
