module Metropolis (MutationScheme(..), metropolis_samples) where

import Algebra
import Math
import Rand

data (RandomGenD g, NormVector v) => MutationScheme g d v extra =
		MutS
		{	muts_eval :: d -> (v, extra)
		,	muts_mutate :: d -> Rand g (d, Double)
		,	muts_initial :: Rand g d
		}

metropolis_samples :: (RandomGenD g, NormVector v) =>
	MutationScheme g d v extra -> Rand g [(v, extra, Bool)]
metropolis_samples muts = muts_initial muts >>= metropolis_samples' muts

metropolis_samples' :: (RandomGenD g, NormVector v) =>
	MutationScheme g d v extra -> d -> Rand g [(v, extra, Bool)]
metropolis_samples' (MutS f mutate _) x0 = bypass (samples x0 ny0 extra_info_0 i0)
	where
		(y0, extra_info_0) = f x0
		i0 = norm y0
		ny0 = y0 ./ i0
		samples x ny extra_info i = mutate x >>= \(x', transfer_ratio) ->
			let	(y', extra_info') = f x'
				i' = norm y'
				ny' = y' ./ i'
				p_accept = min 1 (transfer_ratio * i' / i)
			in
			do	accept <- bias_bool p_accept
				rest <- if accept then samples x' ny' extra_info' i' else samples x ny extra_info i
				return $ (ny .* (1 - p_accept), extra_info, False) : (y' .* (p_accept / (i' + eps)), extra_info', accept) : rest
