module Part (Part(..), Hit (..), EBDRF, End(..), Scene(..), union, point_light, zero_emit, extend_rm_to_ends) where

import Algebra
import Color
import Control.Monad.Trans
import Geometry
import qualified LightProp as LP
import Maybe
import Monad
import Rand
import Types

-- trace returns distance, point and distribution of directions the light can go next
data (RandomGenD g) => Part g =
		Part
		{	trace :: !(Either End Ray -> (MaybeT (Rand g) (Hit g), Double -> LP.T g))
		}

data (RandomGenD g) => Hit g =
		Hit
		{	hit_dist :: Double
		,	hit_loc :: Point
		,	hit_ebdrf :: EBDRF g
		}

type EBDRF g = Either End Vec -> RandMeas g (Either End Vec) (LP.T g)

data End =
		Source
	|	Sink
	deriving Eq

data (RandomGenD g) => Scene g =
		Scene
			{	part_ :: Part g
			}

union :: (RandomGenD g) => [Part g] -> Part g
union parts = Part union_trace where
	union_trace ray@(Left _) = (hit, error "union_trace: dim from source/sink undefined") where
		hit =
			do	i <- lift $ rand_n n
				first_hit <- fst (trace (parts !! i) ray)
				return $ Hit
					(error "union_trace: distance from source/sink undefined")
					(hit_loc first_hit)
					(hit_ebdrf first_hit .* fromIntegral n)

	union_trace ray@(Right _) = (hit, vprod (map snd traces)) where
		hit =
			do	intersections <- lift $ sequence (map (run_maybe_t . fst) traces)
				MaybeT $ return $ closest_intersection_maybe (catMaybes intersections)
		traces = [trace part ray | part <- parts]

	n = length parts

	closest_intersection_maybe [] = Nothing
	closest_intersection_maybe (x : rest) = Just (closest_intersection x rest)
	closest_intersection x [] = x
	closest_intersection x (y : rest) =
		closest_intersection
			(	if hit_dist x <= hit_dist y
				then x
				else y
			)
			rest

point_light :: (RandomGenD g) => Point -> Part g
point_light p = Part trace where
	trace (Left Source) = (hit, idvec) where
		hit = return $
			Hit
				(error "point_light: distance from source undefined")
				p
				bdrf
	trace _ = (MaybeT $ return Nothing, idvec)

	bdrf (Left Source) = extend_rm_to_ends (unif_randmeas rand_norm_vec)

	bdrf _ = zero_emit

zero_emit :: (RandomGenD g) => RandMeas g (Either End Vec) (LP.T g)
zero_emit = unif_randmeas (liftM Right rand_norm_vec) .* 0

extend_rm_to_ends :: (RandomGenD g, Vector v) => RandMeas g d v -> RandMeas g (Either a d) v
extend_rm_to_ends m = RM ac_p ac_f ac_r ac_rf ms_r where
	ac_p = ac_p_ m
	ac_f = either (const zvec) (ac_f_ m)
	ac_r = liftM (\(v, d) -> (v, Right d)) (ac_r_ m)
	ac_rf = either (error "extend_rm_to_ends: ac_rf: undefined on Left") (ac_rf_ m)
	ms_r = liftM (\(v, d) -> (v, Right d)) (ms_r_ m)
