module RListSample (T, muts) where

import Algebra
import CUtil
import Color
import Control.Monad.Trans (lift)
import Geometry
import qualified LightProp as LP
import Math
import Maybe
import Metropolis
import Monad
import Part
import RList
import Rand
import SList
import Types
import Util

data (RandomGenD g) => T g = C !(SList Double) !(RGB, (LP.T (RList g), [Point]))

muts :: (RandomGenD g) => Scene (RList g) -> Double -> Double -> MutationScheme g (T g) RGB (LP.T (RList g), [Point])
muts scene stddev p_large = MutS eval (mutate scene stddev p_large) (initial scene)

eval :: (RandomGenD g) => T g -> (RGB, (LP.T (RList g), [Point]))
eval (C _ v) = v

--tri :: Double -> Double
--tri x = if mod n 2 == 0 then r else 1 - r where (n, r) = properFraction (abs x)

mutate :: (RandomGenD g) => Scene (RList g) -> Double -> Double -> T g -> Rand g (T g, Double)
mutate scene stddev p_large (C l0 _) =
	do	l1' <- mutate' l0
		(lp_stat_1@(lp1, _), l1) <- run_rlist (rand_path scene) l1'
		return (C l1 (LP.color lp1, lp_stat_1), 1)
	where
		mutate' l =
			do	large <- bias_bool p_large
				if large
					then return Empty
					else {-# SCC "RLM_mutate''" #-} mutate'' l
		mutate'' Empty = return Empty
		mutate'' (Cons h t) =
			do	h' <- liftM (tri . (h +) . (stddev *)) rand_norm
				liftM (Cons h') (mutate'' t)

initial :: (RandomGenD g) => Scene (RList g) -> Rand g (T g)
initial scene =
	do	(lp_stat_0@(lp0, _), l0) <- rand_cond ((eps <) . norm . LP.color . fst . fst) (run_rlist (rand_path scene) Empty)
		return (C l0 (LP.color lp0, lp_stat_0))



-- TODO: Make it so the initial and final segments don't have to end at points: they should
-- be able to end immediately, at the source or sink.
rand_path :: (RandomGenD g) => Scene g -> Rand g (LP.T g, [Point])
rand_path scene = liftM (fromMaybe (zvec, [])) $ run_maybe_t m where
	m =
		do	(n_fwd, fwd_base, fwd_dir, fwd_stat) <- trace_path scene Source
			(n_bwd, bwd_base, bwd_dir, bwd_stat) <- trace_path scene Sink
			(dist, distSqr, dir, undir) <- MaybeT $ return $ join_info fwd_base bwd_base
			lp <- return $ (ac_f_ fwd_dir dir @* ac_f_ bwd_dir undir) .* (ac_p_ fwd_dir * ac_p_ bwd_dir)
			dim <- either (const (return idvec)) (\d -> block_or_dim dist (Right (from_right fwd_base, d))) dir
			return (lp @* dim ./ (distSqr * fromIntegral (n_fwd + 1 + n_bwd)), reverse fwd_stat ++ bwd_stat)
	block_or_dim dist ray = let (hit, dim) = trace (part_ scene) ray in
		do	block <- lift $ run_maybe_t $ hit
			case block of
				Nothing -> return (dim dist)
				(Just hit)
					|	hit_dist hit < dist - eps
						-> MaybeT (return Nothing)
					|	otherwise
						-> return (dim dist)

join_info :: Either End Point -> Either End Point -> Maybe (Double, Double, Either End Vec, Either End Vec)
join_info (Left _) (Left _) = Nothing
join_info end@(Left _) (Right _) = Just (1, 1, end, end)
join_info (Right _) end@(Left _) = Just (1, 1, end, end)
join_info (Right u) (Right v) = Just (norm diff, normSqr diff, Right dir, Right (dir .* (- 1)))
	where
		diff = v %- u
		dir = normalize diff

max_half_path_length :: Int
max_half_path_length = 100

trace_path :: (RandomGenD g) =>
		Scene g
	->	End
	->	MaybeT (Rand g) (Int, Either End Point, RandMeas g (Either End Vec) (LP.T g), [Point])
trace_path scene end = trace' [] 0 0 False idvec (Left end) (ac_randmeas (\x -> if either (== end) (const False) x then idvec else zvec) (return (idvec, Left end))) where
	trace' stat n_pair_ac_steps n_steps last_was_ac lp source dir_m =
		let
			p_last = if n_steps == max_half_path_length then 0 else 1 / (fromIntegral (n_steps + 2))
		in
		do	guard $ n_steps <= max_half_path_length
			last_step <- lift $ bias_bool (p_last * ac_p_ dir_m)
			if last_step
				then
					return
					(	n_pair_ac_steps + indic last_was_ac
					,	source
					,	mult_rm dir_m (lp ./ p_last)
					,	stat
					)
				else
				do	(this_was_ac, lp', dir_sample) <- lift $ rm_sample_tell dir_m
					-- not allowed to go to an End
					guard $ either (const (is_left source)) (const True) dir_sample
					(hitm, dim_f) <- return $ trace (part_ scene) (either Left (\r -> Right (r, from_right dir_sample)) source)
					hit <- hitm
					dim <- return $ either idvec (const (dim_f (hit_dist hit))) dir_sample
					dir_m <- return $ hit_ebdrf hit dir_sample
					trace'
						(hit_loc hit : stat)
						(n_pair_ac_steps + indic (last_was_ac && this_was_ac))
						(succ n_steps)
						this_was_ac
						(lp @* lp' @* dim ./ (1 - p_last * ac_p_ dir_m))
						(Right (hit_loc hit))
						dir_m
