module RasterUtil
( readStorableArray
, writeStorableArray
, addRasterTo
, readToPtr
, writeFromPtr
) where

import Data.Array
import Data.Array.MArray
import qualified Data.Array.Storable as StoA
import Data.Monoid
import Data.Word
import Foreign.Ptr
import Foreign.Storable
import Monad
import System.Posix.IO
import System.Posix.Types

readStorableArray :: (Storable a, Monoid a, Ix b) => Ptr a -> (b, b) -> IO (StoA.StorableArray b a)
                     -- Monoid a is just a hack to get a default value mempty
readStorableArray p b = StoA.newArray b mempty >>= \a ->
    let f li i =
            do v <- peekElemOff p li
               writeArray a i v
    in sequence_ [f li i | (li, i) <- zip [0..] (range b)] >> return a

writeStorableArray :: (Storable a, Ix b) => Ptr a -> StoA.StorableArray b a -> IO ()
writeStorableArray p a = getBounds a >>= \b ->
    let f li i =
            do v <- readArray a i
               pokeElemOff p li v
    in sequence_ [f li i | (li, i) <- zip [0..] (range b)]

addRasterTo :: (Storable a, Monoid a, Ix b) => StoA.StorableArray b a -> Array b a -> IO ()
addRasterTo dest src = getBounds dest >>= \destBounds ->
    let f i =
            do v0 <- readArray dest i
               let v1 = src ! i
               writeArray dest i (mappend v0 v1)
    in if destBounds == bounds src
           then sequence_ (map f (range destBounds))
           else fail "addRasterTo: different sizes"

readToPtr :: Fd -> Ptr Word8 -> ByteCount -> IO ()
readToPtr fd ptr bytesWanted =
    do bytesRead <- fdReadBuf fd ptr bytesWanted
       when (bytesWanted /= bytesRead) $ fail "readToPtr: not enough read (TODO: should we read again?)"

writeFromPtr :: Fd -> Ptr Word8 -> ByteCount -> IO ()
writeFromPtr fd ptr bytesWanted =
    do bytesWritten <- fdWriteBuf fd ptr bytesWanted
       when (bytesWanted /= bytesWritten) $ fail "writeFromPtr: not enough written (TODO: should we write again?)"
