module Scenes.Box (scene, eye, gaze, up, right) where

import Algebra
import Applicable
import Camera
import Color
import Geometry
import Material
import Part
import Rand
import Shape

scene :: (RandomGenD g) => Scene g
scene = Scene
	(	union
		[	camera eye gaze up right
		,	point_light (V 0 10 0)
		,	point_light (V (- 5) 3 6)
		,	the_box
		,	the_sphbox
		,	the_floor
		]
	)

the_box :: (RandomGenD g) => Part g
the_box = shape_to_part (pget box) the_box_mat

the_box_mat :: (RandomGenD g) => Material g
the_box_mat = hollow_mat (color_surface (RGB 1 0.4 0.4) (ibdrf_to_surf_mat lambert))

sphbox :: Prep ATr Shape
sphbox = intersection_prep [papply sphere (A (idvec .* 0.8) (V 0.5 (-0.5) 0.5)), box]

the_sphbox :: (RandomGenD g) => Part g
the_sphbox = shape_to_part (pget (papply sphbox (A idvec (V 0 3 0)))) the_sphbox_mat

the_sphbox_mat :: (RandomGenD g) => Material g
the_sphbox_mat = hollow_mat (color_surface (RGB 0 0 1) (ibdrf_to_surf_mat lambert))

the_floor :: (RandomGenD g) => Part g
the_floor = shape_to_part (pget (plane (V 0 1 0) (- 1))) (hollow_mat (checker (tile zvec) (tile idvec)))

tile :: (RandomGenD g) => Color -> SurfaceMat g
tile color = ibdrf_to_surf_mat (fuzzy_specular 10 .* (3 * 0.5)) `add_surf_mat` color_surface (color .* 0.5) (ibdrf_to_surf_mat lambert)

cross :: Vec -> Vec -> Vec
cross (V x0 x1 x2) (V y0 y1 y2) = V (x1 * y2 - x2 * y1) (x2 * y0 - x0 * y2) (x0 * y1 - x1 * y0)

eye, gaze, up, right', right :: Vec
eye = V (- 1.5) 3 7.5
gaze = V 0.25 (- 0.25) (- 1) .* 0.7
up = normalize (cross gaze right' .* (- 1))
right' = V 0.5 0 0
right = normalize (cross gaze up)
