module Scenes.ThreeSpheres (scene) where

import Algebra
import Applicable
import Camera
import Color
import Control.Monad.Trans
import Geometry
import qualified LightProp as LP
import Material
import Maybe
import Monad
import Part
import Rand
import Shape
import Types

scatter :: (RandomGenD g) => Double -> Part g
scatter mean = Part scatter_trace
	where
		scatter_trace (Left _) = (MaybeT $ return Nothing, idvec)
		scatter_trace (Right (base, dir)) = (hit, idvec) where
			hit = lift $
				do	dist <- liftM (mean *) rand_exp
					return $ Hit dist (base @+ dir .* dist) (const (extend_rm_to_ends (unif_randmeas rand_norm_vec .* 0.9)))

swapyz :: LTr
swapyz = L (V 1 0 0) (V 0 0 1) (V 0 1 0)

glow_sphere :: (RandomGenD g) => Double -> ATr -> Part g
glow_sphere sa t = Part trace
	where
		sph = shape_to_part (pget (apply t (sphere :: Prep ATr Shape) :: Prep ATr Shape)) (glow_mat sa)
		trace (Left Source) = (gen, idvec)
		trace (Left Sink) = (mzero, idvec)
		trace ray@(Right _) = Part.trace sph ray
		gen =
			do	v <- lift $ rand_norm_vec
				(p, dir) <- return $ apply t (v, v)
				normal <- return $ normalize dir
				return $ Hit
					(error "glow_sphere: gen: hit dist undefined")
					p
					(gen_ebdrf normal)
		gen_ebdrf normal (Left Source) = fromJust ((ibdrf_to_surf_mat lambert) 0 zvec normal) (Right (normal .* (- 1)))

glow_mat :: (RandomGenD g) => Double -> Material g
glow_mat surface_area = Mat surf_mat idvec idvec 0
	where
		surf_mat _ normal _ = Just (glow_ebdrf normal)
		glow_ebdrf _ (Left _) = zero_emit
		glow_ebdrf normal (Right v) = if normal @. v > 0 then ac_randmeas (\x -> if either (Source ==) (const False) x then idvec else zvec) (return (idvec, Left Source)) else zero_emit

the_floor :: (RandomGenD g) => Part g
the_floor = shape_to_part (pget (plane (V 0 1 0) (- 1))) (hollow_mat (checker (tile zvec) (tile idvec)))

tile :: (RandomGenD g) => Color -> SurfaceMat g
tile color = ibdrf_to_surf_mat (fuzzy_specular 10 .* (3 * 0.5)) `add_surf_mat` color_surface (color .* 0.5) (ibdrf_to_surf_mat lambert)

the_red_sphere :: (RandomGenD g) => Part g
the_red_sphere = shape_to_part (pget (apply (A idvec (V (- 2.1) 0 0)) sphere :: Prep ATr Shape)) the_red_mat

the_red_mat :: (RandomGenD g) => Material g
the_red_mat = hollow_mat (color_surface (RGB 0.25 0 0) (ibdrf_to_surf_mat (lambert @+ fuzzy_specular 3 .* 3)))

the_green_sphere :: (RandomGenD g) => Part g
the_green_sphere = shape_to_part (pget (sphere :: Prep ATr Shape)) the_green_mat

the_green_mat :: (RandomGenD g) => Material g
the_green_mat = hollow_mat (ibdrf_to_surf_mat (ideal_specular .* 0.5) `add_surf_mat` color_surface (RGB 0 0.25 0) (ibdrf_to_surf_mat lambert))

the_blue_sphere :: (RandomGenD g) => Part g
the_blue_sphere = shape_to_part (pget (apply (A idvec (V 2.1 0 0)) sphere :: Prep ATr Shape)) the_blue_mat

the_blue_mat :: (RandomGenD g) => Material g
the_blue_mat = Mat (ibdrf_to_surf_mat fresnel) (\_ dist -> LP.C (RGB (0.5 ** dist) (0.5 ** dist) 1) Nothing) idvec 1.55

the_sky_sphere :: (RandomGenD g) => Part g
the_sky_sphere = shape_to_part (pget (apply (A (idvec .* 100) zvec) sphere :: Prep ATr Shape)) the_sky_mat

the_sky_mat :: (RandomGenD g) => Material g
the_sky_mat = hollow_mat (color_surface (RGB 0.8 0.8 1) (ibdrf_to_surf_mat lambert))

scene :: (RandomGenD g) => Scene g
scene = Scene
	(	union
		[	camera eye gaze up right
		,	the_floor
		,	the_red_sphere
		,	the_green_sphere
		,	the_blue_sphere
		,	the_sky_sphere
		,	glow_sphere (4 * pi) (A idvec (V 0 10 0))
		,	glow_sphere (4 * pi) (A idvec (V (- 10) 10 0))
		,	glow_sphere (4 * pi) (A idvec (V 0 10 10))
		]
	)


cross :: Vec -> Vec -> Vec
cross (V x0 x1 x2) (V y0 y1 y2) = V (x1 * y2 - x2 * y1) (x2 * y0 - x0 * y2) (x0 * y1 - x1 * y0)

eye, gaze, up, right :: Vec
eye = V 0 3 7.5
gaze = normalize (V 0 (- 0.25) (- 1))
up = cross gaze right
right = V 0.5 0 0
