module Text.SExpression.Format
(   SExprFormat(..)
,   Field(..)
,   readFormatted
,   showFormatted
) where

import Control.Monad
import qualified Data.Map as Map
import Data.Maybe
import Text.SExpression

class SExprFormat a where
    sExprFormat :: (Int, [(String, Field a)])
    getInitialFields :: a -> [Value]
    consFromFormat :: Monad m => [Value] -> (Map.Map String [Value]) -> m a

data Field a =
    Field
    { fOptional :: Bool
    , fGet :: a -> Maybe [Value]
    }

readFormatted :: (SExprFormat a, Monad m) => [Value] -> m a
readFormatted ss =
    let (nInitialFields, fields) = sExprFormat
        fieldMap = Map.fromList fields
    in
    do when (length ss < nInitialFields) $ fail "readFormatted: missing initial fields"
       let (initialFields, attribs) = splitAt nInitialFields ss
       let f acc [] =
               do sequence_ [unless (fOptional f || Map.member n acc) $ fail ("readFormatted: field missing: " ++ show n)| (n, f) <- fields]
                  sequence_ [unless (Map.member n fieldMap) $ fail ("readFormatted: unknown field: " ++ show n)| (n, _f) <- Map.toList acc]
                  consFromFormat initialFields acc
           f acc (List (Atom fieldName : fieldSs) : attribTail) =
               do when (Map.member fieldName acc) $ fail ("readFormatted: duplicate field: " ++ show fieldName)
                  f (Map.insert fieldName fieldSs acc) attribTail
           f _acc (attribHead : _) = fail ("readFormatted: bad attribute format: " ++ show attribHead)
       ret <- f Map.empty attribs
       let _forceFieldMapType = ignore (fGet (head (Map.elems fieldMap)) ret)
       return ret

showFormatted :: SExprFormat a => a -> [Value]
showFormatted x =
    let (_nInitialFields, fields) = sExprFormat
    in getInitialFields x ++ [List (Atom n : ss) | (n, f) <- fields, ss <- maybeToList (fGet f x)]

ignore :: a -> ()
ignore _ = ()
