module Types (MaybeT(MaybeT, run_maybe_t), retry, LongAccum, doubleToLA, laToDouble) where

import Data.Monoid
import Data.Word
import Foreign.Ptr
import Foreign.Storable
import Maybe
import Monad

import qualified Control.Monad.Trans as MonadT

newtype MaybeT m a = MaybeT { run_maybe_t :: m (Maybe a) }

instance MonadT.MonadTrans MaybeT where
	lift = MaybeT . liftM return

instance (Monad m) => Monad (MaybeT m) where
	return = MaybeT . return . return
	(MaybeT x) >>= f = MaybeT m
		where
			m = do
				u <- x
				case u of
					Nothing -> return Nothing
					Just a -> run_maybe_t (f a)

instance (Monad m) => MonadPlus (MaybeT m) where
	mzero = MaybeT (return Nothing)
	mplus a b = MaybeT $ run_maybe_t a >>= (fromMaybe (run_maybe_t b) . liftM (return . Just))

retry :: (Monad m) => MaybeT m a -> m a
retry m =
	do	attempt <- run_maybe_t m
		fromMaybe (retry m) (liftM return attempt)

data LongAccum = LA !Word64 !Word32

instance Monoid LongAccum where
    mempty = LA 0 0
    mappend (LA h0 l0) (LA h1 l1) =
        let l' = fromIntegral l0 + fromIntegral l1
        in
        LA (h0 + h1 + if l' > fromIntegral (maxBound :: Word32) then 1 else 0) (fromIntegral l')

instance Storable LongAccum where
    -- We assume that the alignment of a Word32 is less than the alignment of a
    -- Word64 and also less than the size of a Word64.
    sizeOf _ = sizeOf (undefined :: Word64) + sizeOf (undefined :: Word32)
    alignment _ = alignment (undefined :: Word64)
    peek address =
        do hi <- peek (castPtr address)
           lo <- peekByteOff (castPtr address) (sizeOf  hi)
           return (LA hi lo)
    poke address (LA hi lo) =
        do poke (castPtr address) hi
           pokeByteOff (castPtr address) (sizeOf hi) lo

doubleToLA :: Double -> LongAccum
doubleToLA d =
    let (intPart, fracPart) = properFraction d
    in LA intPart (floor (fracPart * (fromIntegral (maxBound :: Word32) + 1)))

laToDouble :: LongAccum -> Double
laToDouble (LA hi lo) = (fromIntegral hi + fromIntegral lo / (fromIntegral (maxBound :: Word32) + 1))
