module Util (indic, is_left, is_right, from_left, from_right, merge_by, merge_many_by, readUTF8File) where

import qualified Data.ByteString as BS
import qualified Data.String.UTF8 as UTF8
import IO
import Monad

indic :: (Num a) => Bool -> a
indic False = 0
indic True = 1

is_left :: Either a b -> Bool
is_left (Left _) = True
is_left (Right _) = False

is_right :: Either a b -> Bool
is_right = not . is_left

from_left :: Either a b -> a
from_left (Left x) = x

from_right :: Either a b -> b
from_right (Right x) = x

merge_by :: (a -> a -> Ordering) -> [a] -> [a] -> [a]
merge_by _ [] y = y
merge_by _ x [] = x
merge_by order x@(xh : xt) y@(yh : yt) = case order xh yh of
	LT -> xh : merge_by order xt y
	EQ -> xh : merge_by order xt y
	GT -> yh : merge_by order x yt

merge_many_by :: (a -> a -> Ordering) -> [[a]] -> [a]
merge_many_by order = foldl (merge_by order) []

readUTF8File :: String -> IO String
readUTF8File fileName =
    do handle <- openFile fileName ReadMode
       contents <- liftM (UTF8.toString . UTF8.fromRep) $ BS.hGetContents handle
       seq (length contents) $ hClose handle
       return contents
