module Main where

import Array
import CUtil
import Color
import Control.Monad.Error
import Control.Monad.Identity
import Data.Array.MArray
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable
import IO
import Image
import JobDescription
import RasterUtil
import System
import System.Posix.IO
import Util

main :: IO ()
main =
    do hPutStrLn stderr "reading job description..."
       jobDescStr <- readUTF8File "job_desc"
       jobDesc <- case runErrorT (readDesc jobDescStr)  of
           Identity (Right x) -> return x
           Identity (Left err) -> do hPutStrLn stderr ("error parsing the job description: " ++ err)
                                     exitFailure
       let rasterSize = imageWidth jobDesc * imageHeight jobDesc * max (sizeOf (undefined :: LongRGB)) (alignment (undefined :: LongRGB))
       hPutStrLn stderr "aquiring lock..."
       lockFd <- openFd "lock" ReadOnly Nothing defaultFileFlags
       lockFile lockFd
       hPutStrLn stderr "reading raster..."
       rasterFd <- openFd "raster" ReadOnly Nothing defaultFileFlags
       fd <- openFd "raster" ReadOnly Nothing defaultFileFlags
       ptr <- mallocBytes rasterSize
       readToPtr fd ptr (fromIntegral rasterSize)
       closeFd fd
       hPutStrLn stderr "releasing lock..."
       unlockFile lockFd
       closeFd lockFd
       hPutStrLn stderr "writing PPM..."
       raster :: Array (Int, Int) LongRGB <- readStorableArray (castPtr ptr) ((0, 0), (imageWidth jobDesc - 1, imageHeight jobDesc - 1)) >>= freeze
       free ptr
       write_PPM stdout (longImageToImage raster)
       hPutStrLn stderr "success"
